# 02-Thermal-noise-mp.py.     I. Marti-Vidal.      IMPRS Blackboard Lectures 2011.
# SIMULATES MANY NOISE IMAGES AND CONVOLVES THEM USING DIFFERENT BEAMS. FOR EACH BEAM, 
# COMPUTES THE DYNAMIC RANGE (PEAK/RMS) AND PLOTS THE RESULTING STATISTICS.
# THIS CODE IS PARALLELIZED (I.E., IT RUNS BETTER IN A MULTI-CORE CPU).

import numpy as np
import pylab as pl
from multiprocessing import Pool, Value
import time, sys

# Parameters:
######################################################################
npix =  512       # Size of image (pixels)
noise = 1.0       # Std. of input noise image
nmonte = 10       # Number of simulations per beamsize.
beams = [1,100,1] # minimum, maximum, and step for beam sizes (pixels).
######################################################################


# Function to simulate the noise images (different set of noise images for each beam):

def simulate(beam):

 global nproc, plot

# Initiate things:
 auxrms = 0.0 ; auxdynr = 0.0
 Taper = np.exp(-Dist/(8.*(npix/float(beam))**2.)) #, since 8=2*(2**2) and sigma=beam/2

 if plot.value == 0: # "plot" controls whether to compute statistics by Monte Carlo 
                     #  or to save a noise image.

  for j in range(nmonte):

# Compute the sky image (convolution made by tapering in Fourier space):
   Fourmap = np.random.normal(0.0,noise,(npix,npix))
   Sky1 = np.real(np.fft.ifft2(np.fft.fft2(Fourmap)*Taper))
# Add to statistics:
   Skystats = [np.max(Sky1),np.std(Sky1)]
   auxrms = (auxrms*float(j)+Skystats[1])/(float(j+1))
   auxdynr = (auxdynr*float(j)+Skystats[0]/Skystats[1])/(float(j+1))

# Update the counter and return:
  nproc.value +=1; 
  sys.stdout.write("\r " + str(nproc.value) + " beams processed.")
  sys.stdout.flush()
  return [2.*beam,auxrms,auxdynr]

 else:  # If plot != 0, save a noise image

  Fourmap = np.random.normal(0.0,noise,(npix,npix))
  Sky1 = np.real(np.fft.ifft2(np.fft.fft2(Fourmap)*Taper))
  pl.imshow(Sky1); pl.title("Map with beamsize "+str(beam))
  pl.xlabel("Relative RA (pixels)"); pl.ylabel("Relative Dec (pixels)")
  pl.colorbar(); pl.axis([0,npix,0,npix]); pl.savefig("./map-"+str(beam)+".png");
  pl.clf()
   

if __name__ == '__main__':

#Initiate things:
 global nproc ; nproc = Value('i',0)
 global plot ; plot = Value('i',0)

 allbeams = np.arange(beams[0],beams[1],beams[2]); 

# Compute a matrix to store the distance^2 to the image center 
# (useful to compute the Taper matrix inside the "simulate" function):
 Qr = np.array([i for i in range(-npix/2,0)]+[i for i in range(1,npix/2+1)]); Q2 = Qr*Qr; 
 Dist = np.fft.fftshift(np.outer(Q2,np.ones(npix)) + np.outer(np.ones(npix),Q2))

# Create a multi-processing pool (I've got a dual core, he he! :D):
 pp = Pool()

# Compute the statistics of all the images (trivial parallelization):
 tic = time.time()
 stats = np.array(pp.map(simulate,allbeams))

# Save some example noise images:
 plot.value = 1; 
 mb1 = int(len(allbeams)/3) ; mb2 = int(2*len(allbeams)/3) 
 pp.map(simulate,[allbeams[0],allbeams[mb1],allbeams[mb2],allbeams[-1]])
 pp.close(); pp.join();

# Total processing time:
 tac = time.time()
 print "\n Process duration: ", tac - tic," seconds."

# Plot the statistics:
 pl.title("Law of Large Numbers in an image")
 pl.loglog(stats[:,0],stats[:,1]*stats[:,0],'.b',label="RMS*beam")
 pl.loglog(stats[:,0],stats[:,1],'.r',label="RMS")
 pl.xlabel("Beam size (pixels)");  pl.ylabel("RMS (sigma)")
 pl.legend(numpoints=1)
 
 fig=pl.figure() ; 
 pl.title("Expected SNR in a pure-noise image (size "+str(npix)+"x"+str(npix)+"pix.)")
 pl.plot(stats[:,0],stats[:,2],'.r')
 pl.xlabel("Beam size (pixels)");  pl.ylabel("Dyn. Range")
 pl.show()

# END
